import pandas as pd
import numpy as np
import scanpy as sc

from scipy.sparse import find, csr_matrix
from sklearn.neighbors import NearestNeighbors
from sklearn.linear_model import LinearRegression

from harmony import utils


def augmented_affinity_matrix(data_df, timepoints, timepoint_connections,
                 n_neighbors=30, n_jobs=-2, pc_components=1000):
    """Function for max min sampling of waypoints

    :param data_df: Normalized data frame. Data frame should be sorted according to the timepoints
    :param timepoints: Panadas series indicating timepoints for each cell in data_df
    :param timepoint_connections: Links between timepoints
    :param n_neighbors: Number of nearest neighbors for graph construction
    :param pc_components: Minimum number of principal components to use. Specify `None` to use pre-computed components
    :return: Affinity matrix  augmented to mutually nearest neighbors
    """

    # Timepoints nad data_df should in same order
    timepoints = timepoints[data_df.index]
    cell_order = data_df.index

    # Time point cells and indices
    tp_cells = pd.Series()
    tp_offset = pd.Series()
    offset = 0
    for i in timepoints.unique():
        tp_offset[i] = offset
        tp_cells[i] = list(timepoints.index[timepoints == i])
        offset += len(tp_cells[i])

    # Run PCA to denoise the dropouts
    if pc_components is None:
        pca_projections = data_df
    else:
        pca_projections, _ = utils.run_pca(data_df, n_components=pc_components)

    # Nearest neighbor graph construction and affinity matrix
    print('Nearest neighbor computation...')

    # --------------------------------------------------------------------------
    # nbrs = NearestNeighbors(n_neighbors=n_neighbors,
    #                         metric='euclidean', n_jobs=-2)
    # nbrs.fit(pca_projections.values)
    # dists, _ = nbrs.kneighbors(pca_projections.values)
    # adj = nbrs.kneighbors_graph(pca_projections.values, mode='distance')
    # # Scaling factors for affinity matrix construction
    # ka = np.int(n_neighbors / 3)
    # scaling_factors = pd.Series(dists[:, ka], index=cell_order)
    # # Affinity matrix
    # nn_aff = _convert_to_affinity(adj, scaling_factors, True)
    # --------------------------------------------------------------------------

    temp = sc.AnnData(data_df.values)
    sc.pp.neighbors(temp, n_pcs=0, n_neighbors=n_neighbors)
    kNN = temp.uns['neighbors']['distances']

    # Adaptive k
    adaptive_k = int(np.floor(n_neighbors / 3))
    scaling_factors = np.zeros(data_df.shape[0])

    for i in np.arange(len(scaling_factors)):
        scaling_factors[i] = np.sort(kNN.data[kNN.indptr[i]:kNN.indptr[i + 1]])[adaptive_k - 1]

    scaling_factors = pd.Series(scaling_factors, index=cell_order)

    # Affinity matrix
    nn_aff = _convert_to_affinity(kNN, scaling_factors, True)

    # Mututally nearest neighbor affinity matrix
    # Initilze mnn affinity matrix
    N = len(cell_order)
    full_mnn_aff = csr_matrix(([0], ([0], [0])), [N, N])
    for i in timepoint_connections.index:
        t1, t2 = timepoint_connections.loc[i, :].values
        print(f'Constucting affinities between {t1} and {t2}...')

        # MNN matrix  and distance to ka the distance
        t1_cells = tp_cells[t1]
        t2_cells = tp_cells[t2]
        mnn = _construct_mnn(t1_cells, t2_cells, pca_projections,
                             n_neighbors, n_jobs)

        # MNN Scaling factors
        # Distance to the adaptive neighbor
        ka_dists = pd.Series(0.0, index=t1_cells + t2_cells)
        # T1 scaling factors
        ka_dists[t1_cells] = _mnn_ka_distances(mnn, n_neighbors)
        # T2 scaling factors
        ka_dists[t2_cells] = _mnn_ka_distances(mnn.T, n_neighbors)

        # Scaling factors
        mnn_scaling_factors = pd.Series(0.0, index=cell_order)
        mnn_scaling_factors[t1_cells] = _mnn_scaling_factors(
            ka_dists[t1_cells], scaling_factors)
        mnn_scaling_factors[t2_cells] = _mnn_scaling_factors(
            ka_dists[t2_cells], scaling_factors)

        # MNN affinity matrix
        full_mnn_aff = full_mnn_aff + \
            _mnn_affinity(mnn, mnn_scaling_factors,
                          tp_offset[t1], tp_offset[t2])

    # Symmetrize the affinity matrix and return
    aff = nn_aff + nn_aff.T + full_mnn_aff + full_mnn_aff.T
    return aff, nn_aff + nn_aff.T


def _convert_to_affinity(adj, scaling_factors, with_self_loops=False):
    """ Convert adjacency matrix to affinity matrix
    """
    N = adj.shape[0]
    rows, cols, dists = find(adj)
    dists = dists ** 2/(scaling_factors.values[rows] ** 2)

    # Self loops
    if with_self_loops:
        dists = np.append(dists, np.zeros(N))
        rows = np.append(rows, range(N))
        cols = np.append(cols, range(N))
    aff = csr_matrix((np.exp(-dists), (rows, cols)), shape=[N, N])
    return aff


def _construct_mnn(t1_cells, t2_cells, data_df, n_neighbors, n_jobs=-2):
    # FUnction to construct mutually nearest neighbors bewteen two points

    print(f't+1 neighbors of t...')
    nbrs = NearestNeighbors(n_neighbors=n_neighbors,
                            metric='euclidean', n_jobs=n_jobs)
    nbrs.fit(data_df.loc[t1_cells, :].values)
    t1_nbrs = nbrs.kneighbors_graph(
        data_df.loc[t2_cells, :].values, mode='distance')

    print(f't neighbors of t+1...')
    nbrs = NearestNeighbors(n_neighbors=n_neighbors,
                            metric='euclidean', n_jobs=n_jobs)
    nbrs.fit(data_df.loc[t2_cells, :].values)
    t2_nbrs = nbrs.kneighbors_graph(
        data_df.loc[t1_cells, :].values, mode='distance')

    # Mututally nearest neighbors
    mnn = t2_nbrs.multiply(t1_nbrs.T)
    mnn = mnn.sqrt()
    return mnn


def _mnn_ka_distances(mnn, n_neighbors):
    # Function to find distance ka^th neighbor in the mutual nearest neighbor matrix
    ka = np.int(n_neighbors / 3)
    ka_dists = np.repeat(None, mnn.shape[0])
    for i in range(mnn.shape[0]):
        x, y, z = find(mnn[i, :])
        if len(z) >= ka:
            ka_dists[i] = np.sort(z)[ka - 1]
    return ka_dists


def _mnn_scaling_factors(mnn_ka_dists, scaling_factors):
    cells = mnn_ka_dists.index[~mnn_ka_dists.isnull()]

    # Linear model fit
    x = scaling_factors[cells]
    y = mnn_ka_dists[cells]
    lm = LinearRegression()
    lm.fit(x.values.reshape(-1, 1), y.values.reshape(-1, 1))

    # Predict
    x = scaling_factors[mnn_ka_dists.index]
    vals = np.ravel(lm.predict(x.values.reshape(-1, 1)))
    mnn_scaling_factors = pd.Series(vals, index=mnn_ka_dists.index)

    return mnn_scaling_factors


def _mnn_affinity(mnn, mnn_scaling_factors, offset1, offset2):
    # Function to convert mnn matrix to affinicty matrix

    # Construct adjacency matrix
    N = len(mnn_scaling_factors)
    x, y, z = find(mnn)
    x = x + offset1
    y = y + offset2
    adj = csr_matrix((z, (x, y)), shape=[N, N])

    # Affinity matrix
    return _convert_to_affinity(adj, mnn_scaling_factors, False)



import sys
from setuptools import setup
from warnings import warn

if sys.version_info.major != 3:
    raise RuntimeError("Palantir requires Python 3")
if sys.version_info.minor < 6:
    warn("Analysis methods were developed using Python 3.6")

# get version
with open("src/harmony/version.py") as f:
    exec(f.read())
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="harmonyTS",
    version=__version__,  # read in from the exec of version.py; ignore error
    description=(
        "Harmony is a unified framework for data visualization, analysis "
        "and interpretation of scRNA-seq data measured across discrete time points"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dpeerlab/harmony",
    author="Manu Setty",
    author_email="manu.talanki@gmail.com",
    package_dir={"": "src"},
    packages=["harmony"],
    install_requires=[
        "numpy>=1.14.2",
        "pandas>=0.22.0",
        "scipy>=1.0.1",
        "sklearn",
        "fa2",
        "matplotlib>=2.2.2",
        "seaborn>=0.8.1",
        "palantir>=0.2.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    python_requires='>=3.6',
)



import warnings
import os
import numpy as np
import pandas as pd
import copy
from itertools import chain
from sklearn.preprocessing import StandardScaler

import matplotlib
from matplotlib import font_manager

try:
    os.environ['DISPLAY']
except KeyError:
    matplotlib.use('Agg')

import matplotlib.pyplot as plt

with warnings.catch_warnings():
    # catch experimental ipython widget warning
    warnings.simplefilter('ignore')
    import seaborn as sns
    sns.set(context="paper", style='ticks',
            font_scale=1.5, font='Bitstream Vera Sans')

# set plotting defaults
with warnings.catch_warnings():
    # catch warnings that system can't find fonts
    warnings.simplefilter('ignore')
    import seaborn as sns
    fm = font_manager.fontManager
    fm.findfont('Raleway')
    fm.findfont('Lato')

warnings.filterwarnings(action="ignore", message="remove_na is deprecated")

from fa2 import ForceAtlas2
import pandas as pd
import numpy as np
import random

def force_directed_layout(affinity_matrix, cell_names=None, verbose=True, iterations=500):
    """" Function to compute force directed layout from the affinity_matrix
    :param affinity_matrix: Sparse matrix representing affinities between cells
    :param cell_names: pandas Series object with cell names
    :param verbose: Verbosity for force directed layout computation
    :param iterations: Number of iterations used by ForceAtlas 
    :return: Pandas data frame representing the force directed layout
    """

    init_coords = np.random.random((affinity_matrix.shape[0], 2))

    forceatlas2 = ForceAtlas2(
        # Behavior alternatives
        outboundAttractionDistribution=False,  
        linLogMode=False,  
        adjustSizes=False,  
        edgeWeightInfluence=1.0,
        # Performance
        jitterTolerance=1.0,  
        barnesHutOptimize=True,
        barnesHutTheta=1.2,
        multiThreaded=False,  
        # Tuning
        scalingRatio=2.0,
        strongGravityMode=False,
        gravity=1.0,
        # Log
        verbose=verbose)

    positions = forceatlas2.forceatlas2(
        affinity_matrix, pos=init_coords, iterations=iterations)
    positions = np.array(positions)

    # Convert to dataframe
    if cell_names is None:
        cell_names = np.arange(affinity_matrix.shape[0])

    positions = pd.DataFrame(positions,
                             index=cell_names, columns=['x', 'y'])
    return positions


class FigureGrid:
    """
    Generates a grid of axes for plotting
    axes can be iterated over or selected by number. e.g.:
    >>> # iterate over axes and plot some nonsense
    >>> fig = FigureGrid(4, max_cols=2)
    >>> for i, ax in enumerate(fig):
    >>>     plt.plot(np.arange(10) * i)
    >>> # select axis using indexing
    >>> ax3 = fig[3]
    >>> ax3.set_title("I'm axis 3")
    """

    # Figure Grid is favorable for displaying multiple graphs side by side.

    def __init__(self, n: int, max_cols=3, scale=3):
        """
        :param n: number of axes to generate
        :param max_cols: maximum number of axes in a given row
        """

        self.n = n
        self.nrows = int(np.ceil(n / max_cols))
        self.ncols = int(min((max_cols, n)))
        figsize = self.ncols * scale, self.nrows * scale

        # create figure
        self.gs = plt.GridSpec(nrows=self.nrows, ncols=self.ncols)
        self.figure = plt.figure(figsize=figsize)

        # create axes
        self.axes = {}
        for i in range(n):
            row = int(i // self.ncols)
            col = int(i % self.ncols)
            self.axes[i] = plt.subplot(self.gs[row, col])

    def __getitem__(self, item):
        return self.axes[item]

    def __iter__(self):
        for i in range(self.n):
            yield self[i]


def plot_timepoints(layout, timepoints):
    """Plot timepoints on the force directed layoug
    :param layout: Force directed layout
    :param timepoints: Pandas series of timepoints
    """

    # Cluster colors
    n_clusters = len(set(timepoints))
    cluster_colors = pd.Series(sns.color_palette(
        'muted', n_clusters).as_hex(), index=np.sort(timepoints.unique()))

    # Set up figure
    fig = FigureGrid(n_clusters, 5)
    for t, ax in zip(cluster_colors.index, fig):
        ax.scatter(layout['x'], layout['y'], s=3, color='lightgrey')
        cells = timepoints.index[timepoints == t]
        ax.scatter(layout.loc[cells, 'x'], layout.loc[cells, 'y'], 
            s=5, color=cluster_colors[t])
        ax.set_axis_off()
        ax.set_title(t)


def plot_tp_gene_expression(data, layout, genes, timepoints):
    """ Plot gene expression on force directed layout
    :param genes: Iterable of strings to plot on force directed layout
    """

    not_in_dataframe = set(genes).difference(data.columns)
    if not_in_dataframe:
        if len(not_in_dataframe) < len(genes):
            print('The following genes were either not observed in the experiment, '
                  'or the wrong gene symbol was used: {!r}'.format(not_in_dataframe))
        else:
            print('None of the listed genes were observed in the experiment, or the '
                  'wrong symbols were used.')
            return

    # remove genes missing from experiment
    genes = set(genes).difference(not_in_dataframe)

    # Plot
    ncols = len(genes); nrows = len(timepoints.unique());
    fig = plt.figure(figsize=[4 * ncols, 4 * nrows])
    gs = plt.GridSpec(nrows, ncols)

    for i, t in enumerate(np.sort(timepoints.unique())):
        cells = timepoints.index[timepoints == t]

        for j, gene in enumerate(genes):
            c = data.loc[:, gene]

            ax = plt.subplot(gs[i, j])
            ax.scatter(layout.loc[:, 'x'], layout.loc[:, 'y'], s=3, color='lightgrey')
            ax.scatter(layout.loc[cells, 'x'], layout.loc[cells, 'y'], s=5,
                cmap=matplotlib.cm.Spectral_r, c=c[cells], vmin=c.min(), vmax=c.max())
            ax.set_title(f'{t}: {gene}')
            ax.set_axis_off()





import pandas as pd
import numpy as np
import os
import re
from itertools import chain
from collections import OrderedDict
import warnings
from sklearn.decomposition import PCA


def load_from_csvs(csv_files, sample_names=None, min_cell_count=10):
    """ Read in scRNA-seq data from csv files
    :param csv_files: List of csv files
    :param sample_names: Prefix to attach to the cell names
    :param min_cell_count: Minimum number of cells a gene should be detected in
    :return: Pandas data frame representing the count matrix
    """
    if sample_names is None:
        sample_names = [re.sub('.csv', '', os.path.split(i)[-1]) for i in csv_files]


    # Load counts
    print('Loading count matrices...')
    counts_dict = OrderedDict()

    for csv_file, sample in zip(csv_files, sample_names):
        print(sample)
        counts = pd.read_csv(csv_file, index_col=0)

        # Update sample names
        counts.index = sample + '_' + counts.index.astype(str)

        # Remove zero count genes
        counts = counts.loc[:, counts.sum() > 0]
        counts = counts.astype(np.int16)

        # Update dictionary
        counts_dict[sample] = counts

    # Concatenate cells and genes
    print('Concatenating data..')
    all_cells = list(chain(*[list(counts_dict[sample].index)
                             for sample in sample_names]))
    all_genes = list(
        chain(*[list(counts_dict[sample].columns) for sample in sample_names]))
    all_genes = list(set(all_genes))

    # Counts matrix
    counts = pd.DataFrame(0, index=all_cells,
                          columns=all_genes, dtype=np.int16)
    for sample in sample_names:
        sample_counts = counts_dict[sample]
        counts.loc[sample_counts.index, sample_counts.columns] = sample_counts

    # Filter out low detection genes
    gs = counts.sum()
    counts = counts.loc[:, counts.columns[gs > min_cell_count]]

    return counts


def hvg_genes(norm_df, no_genes=1000):
    """ Select genes based on normalized dispersion
    """

    # Mean and variance
    mean = norm_df.mean()
    var = norm_df.var()

    # Construct data frame
    df = pd.DataFrame()
    df['mean'] = mean
    df['dispersion'] = var/mean
    df[df.isnull()] = 0

    # Normalized disperion
    from statsmodels import robust
    df['mean_bin'] = pd.cut(df['mean'], np.r_[-np.inf,
                                              np.percentile(df['mean'],
                                                            np.arange(10, 105, 5)), np.inf])
    disp_grouped = df.groupby('mean_bin')['dispersion']
    disp_median_bin = disp_grouped.median()
    # the next line raises the warning: "Mean of empty slice"
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        disp_mad_bin = disp_grouped.apply(robust.mad)
    df['dispersion_norm'] = np.abs((df['dispersion'].values - disp_median_bin[df['mean_bin'].values].values)) \
        / disp_mad_bin[df['mean_bin'].values].values

    # Subset of genes
    use_genes = df['dispersion_norm'].sort_values().index[::-1][:no_genes]

    return use_genes


def run_pca(data, n_components=300, var_explained=0.85):
    """Run PCA

    :param data: Dataframe of cells X genes. Typicaly multiscale space diffusion components
    :param n_components: Number of principal components
    :param var_explained: Include components that explain amount variance. Note
    number of components = min(n_components, components explaining var_explained)
    :return: PCA projections of the data and the explained variance
    """
    init_components = min([n_components, data.shape[0]])
    pca = PCA(n_components=init_components, svd_solver='randomized')
    pca.fit(data)
    if pca.explained_variance_ratio_.sum() >= 0.85:
        n_components = np.where(np.cumsum(pca.explained_variance_ratio_) >= var_explained)[0][0]

    print(f'Running PCA with {n_components} components')
    pca = PCA(n_components=n_components, svd_solver='randomized')
    pca_projections = pca.fit_transform(data)
    pca_projections = pd.DataFrame(pca_projections, index=data.index)
    return pca_projections, pca.explained_variance_ratio_


def normalize_counts(data, multiplier=10000):
    """Correct the counts for molecule count variability

    :param data: Counts matrix: Cells x Genes
    :return: Normalized matrix
    """
    ms = data.sum(axis=1)
    norm_df = data.div(ms, axis=0) * multiplier
    return norm_df


def log_transform(data, pseudo_count=0.1):
    """Log transform the matrix

    :param data: Counts matrix: Cells x Genes
    :return: Log transformed matrix
    """
    return np.log2(data + pseudo_count)
